;; Generated package description from excorporate.el
(define-package "excorporate" "0.7.3" "Exchange integration" '((emacs "24.1") (fsm "0.2") (soap-client "3.1.1") (url-http-ntlm "2.0.2")) :url "http://elpa.gnu.org/packages/excorporate.html" :keywords '("calendar"))
